/*
 * Web site error daemon - configuration file parser
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "library.h"
#include "log.h"
#include "hash.h"
#include "smalloc.h"
#include "regexlist.h"
#include "servicehash.h"
#include "runtime.h"
#include "dsyslog.h"

#define F_INVERT	(1)	/* Invert meaning of clause */
#define F_SINGLE	(2)	/* Single priority only */

#ifdef DPRINTF
#undef DPRINTF
#define DPRINTF(x) do { } while(0)
#endif

#include "pri.c"

static int
dsyslog_getpri(char *str)
{
	const struct _pripair *pair;
	pair = dsyslog_pripair(str, strlen(str));
	if (!pair) return PRI_UNKNOWN;
	return pair->priority;
}

static void
dsyslog_setpri(hash_t	*shash,
				char	*key,
				char	*data,
				int		 pri,
				int		 val)
{
	DPRINTF((stderr, "%s.%d set %s to %d\n", key, pri, data, val));
	servicehash_set(shash, key, data, pri, val);
}

static void
dsyslog_makepri(hash_t	*shash,
				char	*key,
				char	*data,
				int		 priority,
				char	 flags)
{
	int		 i;
	int		 val;

	if (priority == PRI_ALL) {
		for (i = 0; i < PRI_NPRI; i++)
			dsyslog_setpri(shash, key, data, i, 1);
	}
	else if (priority == PRI_NONE) {
		for (i = 0; i < PRI_NPRI; i++)
			dsyslog_setpri(shash, key, data, i, 0);
	}
	else {
		val = !(flags&F_INVERT);	/* Hehehehe */

		if (flags&F_SINGLE) {
			dsyslog_setpri(shash, key, data, priority, val);
		}
		else {
			for (i = 0; i < priority; i++)
				dsyslog_setpri(shash, key, data, i, val);
		}
	}
}

hash_t *
dsyslog_shash(char *select, char *data, char *path, int linenum)
{
	hash_t	*shash;
	char	*key;
	char	*value;
	int		 priority;
	int		 flags;

	shash = servicehash_new();
	key = strtok(select, ";");
	while (key) {
		value = strchr(key, '.');
		if (!value) {
			printf("Syntax error in '%s' on line %d: "
							"Expected key.value pair in '%s'\n",
							path, linenum, key);
			servicehash_delete(shash);
			return NULL;		/* One error per line is enough */
		}
		*value++ = 0;

		flags = 0;
		if (*value == '!') {
			flags |= F_INVERT;
			value++;
		}
		if (*value == '=') {
			flags |= F_SINGLE;
			value++;
		}

		priority = dsyslog_getpri(value);
		if (priority == PRI_UNKNOWN) {
			printf("Syntax error in '%s' on line %d: Could not "
							"identify priority part in '%s.%s'\n",
							path, linenum, key, value);
			servicehash_delete(shash);
			return NULL;		/* One error per line is enough */
		}

		dsyslog_makepri(shash, key, data, priority, flags);

		key = strtok(NULL, ";");
	}
	return shash;
}

regexlist_t *
dsyslog_rlist(char *host, char *path, int linenum)
{
	regexlist_t	*rlist;
	char		*next;
	bool		 invert;

	rlist = regexlist_new();
	next = strtok(host, ";");
	while (next) {
		invert = FALSE;
		while (*next == '!') {
			invert = !invert;
			next++;
		}
		regexlist_push_back(rlist, next, invert);
		if (wregex_error()) {
			fprintf(stderr, "Failed to compile expression '%s' in %s "
							"on line %d\n", next, path, linenum);
			fprintf(stderr, "Error was '%s'\n", wregex_errorstr());
			regexlist_delete(rlist);
			return NULL;		/* One error per line is enough */
		}
		next = strtok(NULL, ";");
	}
	return rlist;
}

/* Reads the config file 'path' into the runtime vector r */
void
dsyslog_read(vector_t *r, char *path)
{
	char	 line[BUFSIZ];
	FILE	*input;
	char	*host;		/* A ;-separated list of regexes */
	char	*select;	/* A ;-separated list of facility-priority */
	char	*data;		/* An action to perform */
	int		 linenum;
	int		 i;

	regexlist_t	*rlist;
	hash_t		*shash;

	input = fopen(path, "r");
	if (!input) {
		perror(path);
		exit(-1);
	}

	linenum = 0;

	while (fgets(line, BUFSIZ, input)) {
		linenum++;
		i = strcspn(line, "#\r\n");	/* Cache */
		if (!i)
			continue;
		if (line[0] == '#')
			continue;
		line[i] = 0;

		host = line;

		while (isspace(*host)) host++;
		select = strpbrk(host, " \t");
		if (!select) {
			fprintf(stderr, "Syntax error in '%s' on line %d: "
							"Expected three fields\n", path, linenum);
			continue;
		}
		*select++ = 0;

		while (isspace(*select)) select++;
		data = strpbrk(select, " \t");
		if (!data) {
			fprintf(stderr, "Syntax error in '%s' on line %d: "
							"Expected three fields\n", path, linenum);
			continue;
		}
		*data++ = 0;

		while (isspace(*data)) data++;

		rlist = dsyslog_rlist(host, path, linenum);
		if (!rlist) continue;
		shash = dsyslog_shash(select, data, path, linenum);
		if (!shash) {
			regexlist_delete(rlist);
			continue;
		}

		/* XXX Push it on */
		runtime_push_back(r, rlist, shash);
	}
}
