/*
 * Virtual Card CGI - file I/O routines
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include "library.h"
#include "hash.h"
#include "rope.h"
#include "smalloc.h"
#include "log.h"
#include "cgi.h"

void
file_write(hash_t *cgihash, hash_t *card)
{
	rope_t		*rope;
	char		*hostname;
	char		*instance;
	time_t		 timebuf;
	char		 buf[2048];
	FILE		*output;
	vector_t	*keys;
	char		*key;
	char		*data;
	int			 len;
	int			 i;

	hostname = hash_get(cgihash, ENV_HOSTNAME);
	if (!hostname) {
		log("cgisys", PRI_CRIT, "file_write: Unexpected no "
						ENV_HOSTNAME " in CGI hash table");
		exit(0);
	}
	instance = hash_get(cgihash, "case");
	if (!instance) {
		log("cgisys", PRI_CRIT, "file_write: Unexpected no 'case' in "
						"CGI hash table");
		exit(0);
	}

	time(&timebuf);

	snprintf(buf, 2047, "%.15d", timebuf);
	hash_add(card, "_time", sstrdup(buf));

	/* How we generate the filename is reasonably irrelevant */
	snprintf(buf, 2047, "%.15d%.6d", timebuf, getpid());
	hash_add(cgihash, "card", sstrdup(buf));

	rope = rope_new(DIR_DATA "/");
	rope_add(rope, hostname);
	rope_add(rope, "/vcard/");
	rope_add(rope, instance);
	rope_add(rope, "/cards/");
	rope_add(rope, buf);
	rope_add(rope, ".data");

	output = fopen(rope_chars(rope), "w");
	if (!output) {
		log("cgisys", PRI_CRIT, "Failed to open output file %s : %s",
						rope_chars(rope), strerror(errno));
		exit(0);
	}

	keys = hash_keys(card);
	len = vector_length(keys);
	for (i = 0; i < len; i++) {
		key = vector_data(keys, i);
		if (strcmp(key, "_template") == 0)
			continue;
		data = hash_get(card, key);
		data = cgi_base64_encode(data);
		fprintf(output, "%s %s\n", key, data);
		sfree(data);
	}

	fclose(output);
	rope_delete(rope);
	vector_delete(keys);
}

hash_t *
file_read(hash_t *cgihash)
{
	hash_t	*card;
	rope_t	*rope;
	char	*hostname;
	char	*instance;
	char	*filename;
	FILE	*input;
	char	 line[BUFSIZ];
	char	*data;

	hostname = hash_get(cgihash, ENV_HOSTNAME);
	if (!hostname) {
		log("cgisys", PRI_CRIT, "file_read: Unexpected no "
						ENV_HOSTNAME " in CGI hash table");
		exit(0);
	}
	instance = hash_get(cgihash, "case");
	if (!instance) {
		log("cgisys", PRI_CRIT, "file_read: Unexpected no 'case' in "
						"CGI hash table");
		exit(0);
	}
	filename = hash_get(cgihash, "card");
	if (!filename) {
		log("cgisys", PRI_CRIT, "file_read: Unexpected no 'card' in "
						"card hash table");
		exit(0);
	}

	filename[strcspn(filename, "/")] = 0;

	rope = rope_new(DIR_DATA "/");
	rope_add(rope, hostname);
	rope_add(rope, "/vcard/");
	rope_add(rope, instance);
	rope_add(rope, "/cards/");
	rope_add(rope, filename);
	rope_add(rope, ".data");

	// DPRINTF((stderr, "Attempting to read %s\n", rope_chars(rope)));

	input = fopen(rope_chars(rope), "r");
	if (!input) {
		return NULL;
	}

	card = hash_new(sfree, 0);

	while (fgets(line, BUFSIZ, input)) {
		line[strcspn(line, "\r\n")] = 0;
		data = strchr(line, ' ');
		if (!data) {
			log("cgisys", PRI_WARNING, "Bad line '%s' in file '%s'",
							line, rope_chars(rope));
			continue;
		}

		*data++ = 0;
		hash_add(card, line, cgi_base64_decode(data));
	}

	return card;
}
