/*
 * Web error document for Apache - main routines
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include "library.h"
#include "hash.h"
#include "cgi.h"
#include "log.h"
#include "rope.h"

#define DIR_TEMPLATE DIR_DATA "/"

void template_parse(char *path, char *str, char *cstr, char *hstr);

char *errorcodes[] = {
	[000] = "No code given",	/* Internal use only */

	[400] = "Bad Request",

	[402] = "Payment Required",
	[403] = "Forbidden",
	[404] = "Not Found",
	[405] = "Method Not Allowed",
	[406] = "Not Acceptable",
	[407] = "Proxy Authentication Required",
	[408] = "Request Timeout",
	[409] = "Conflict",
	[410] = "Gone",
	[411] = "Length Required",
	[412] = "Precondition Failed",
	[413] = "Request Entity Too Large",
	[414] = "Request-URI Too Long",
	[415] = "Unsupported Media Type",

	[500] = "Internal Server Error",
	[501] = "Not Implemented",
	[502] = "Bad Gateway",
	[503] = "Service Unavailable",
	[504] = "Gateway Timeout",
	[505] = "HTTP Version Not Supported",
	NULL
};

/* Impossible to get a '../' into this */
void
send_page(char *site, char *status, char *msg, int code)
{
	rope_t	*r;
	char	*httpmessage;

	if (errorcodes[code])
		httpmessage = errorcodes[code];
	else
		httpmessage = "No HTTP error code";

	r = rope_new(DIR_TEMPLATE);

	if (status) {
		rope_add(r, site);
		rope_add(r, "/weberror/templates/");
		rope_add(r, status);
		rope_add(r, ".tpt");
		if (access(rope_chars(r), R_OK) == 0) {
			template_parse(rope_chars(r), msg, status, httpmessage);
			return;
		}
		rope_set(r, DIR_TEMPLATE);
	}

	rope_add(r, site);
	rope_add(r, "/weberror/templates/message.tpt");
	if (access(rope_chars(r), R_OK) == 0) {
		template_parse(rope_chars(r), msg, status, httpmessage);
		return;
	}

	if (status) {
		rope_set(r, DIR_TEMPLATE "/" DEFAULT "/weberror/templates/");
		rope_add(r, status);
		rope_add(r, ".tpt");
		if (access(rope_chars(r), R_OK) == 0) {
			template_parse(rope_chars(r), msg, status, httpmessage);
			return;
		}
	}

	rope_set(r, DIR_TEMPLATE "/" DEFAULT "/weberror/templates/message.tpt");
	if (access(rope_chars(r), R_OK) == 0) {
		template_parse(rope_chars(r), msg, status, httpmessage);
		return;
	}

	log("cgisys", PRI_ERR, "Could not open file \"%s\" - error "
					"was \"%s\"", rope_chars(r), strerror(errno));
	printf("Content-type: text/plain\r\n\r\n");
	printf("An error was caused\r\n");
	printf("The error message was \"%s\"\r\n", httpmessage);
}

/* log.c does all the NULL-termination checking for us in
 * log_boring() */
void
msg_send(char *status, int code)
{
	errmsg_t	 msg;
	char		*site;
	char		*temp;
	char		*message;

	memset(&msg, 0, sizeof(msg));
	
	site = getenv(ENV_HOSTNAME);
	if (!site) {
		temp = "Could not get " ENV_HOSTNAME;
		site = DEFAULT;
	}
	else {
		temp = site;
	}
	strncpy(msg.site, temp, (LOG_SITELEN - 1));

	temp = getenv("HTTP_REFERER");
	if (!temp) temp = "Could not get HTTP_REFERER";
	strncpy(msg.referer, temp, (LOG_REFERERLEN - 1));

	temp = getenv("REDIRECT_URL");
	if (!temp) temp = "Could not get REDIRECT_URL";
	strncpy(msg.url, temp, (LOG_URLLEN - 1));

	message = getenv("REDIRECT_ERROR_NOTES");
	if (!message)
		message = errorcodes[code];
	if (message)
		strncpy(msg.message, message, (LOG_MESSAGELEN - 1));
	else {
		snprintf(msg.message, (LOG_MESSAGELEN - 1),
						"Unknown error code '%d'", code);
	}

	snprintf(msg.facility, (LOG_FACILITYLEN - 1), "error%d", code);
	msg.priority = PRI_ERR;
	
	/* This terminates msg.message by reference */
	log_msg(&msg, LOG_HOST, LOG_PORT);
	/* This sends the template to the user */
	send_page(site, status, msg.message, code);

}

int
main(void)
{
	char	*status;
	int		 code;

	library_init("weberror");

	/* Template name */
	status = getenv("REDIRECT_STATUS");
	status[strcspn(status, "/")] = 0;
	if (!status || (strlen(status) != 3)) {
		log("cgisys", PRI_NOTICE, "Unexpected REDIRECT_STATUS '%s'",
						status);
		status = NULL;
		code = 0;
	}
	else {
		code = atol(status);
	}

	msg_send(status, code);

	return 0;
}

/*
 * REDIRECT_URL
 * REDIRECT_HTTP_ACCEPT
 * REDIRECT_HTTP_USER_AGENT
 * REDIRECT_QUERY_STRING
 * REDIRECT_REMOTE_ADDR
 * REDIRECT_REMOTE_HOST
 * REDIRECT_SERVER_SOFTWARE
 * REDIRECT_SERVER_NAME 
 * REDIRECT_SERVER_PORT
 * REDIRECT_PATH
 * REDIRECT_REFERER_URL
 */
